# -*- coding: utf-8 -*-
"""
要求配置:
    GA_ID : Google分析的跟踪ID
    GA_AGENT_HOST : GAAgent的启动IP
    GA_AGENT_PORT : GAAgent的启动端口
    GA_FORBID_PATHS : 被拒绝的paths，优先级高于 GA_ALLOW_PATHS
    GA_ALLOW_PATHS : 被允许的paths
    GA_HACK_PATHS: 上报时按照规则替换，在allow和forbid判断之后进行. 示例: [(r'^/all(\S+)', r'ok\g<1>')]
    GA_LOGGER_NAME : 用来打印log的name
"""

import logging
import json
import socket
import errno
import time
import re
import urlparse


class GAAdapter(object):
    _ga_id = None
    _ga_agent_host = None
    _ga_agent_port = None
    _ga_forbid_paths = None
    _ga_allow_paths = None
    _ga_hack_paths = None
    _ga_logger_name = None

    _local_ip = ''

    _socket = None

    def __init__(self):
        self._local_ip = socket.gethostbyname(socket.gethostname()) or ''

        self._socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # 非阻塞
        self._socket.setblocking(0)

    def __del__(self):
        if self._socket:
            self._socket.close()

    def send_data_to_ga_agent(self, send_dict):
        """
        可以在网站中调用
        """
        try:
            self._socket.sendto(json.dumps(send_dict), (self._ga_agent_host, self._ga_agent_port))
        except socket.error, e:
            # errno.EWOULDBLOCK = errno.EAGAIN = 11
            if e.args[0] == errno.EWOULDBLOCK:
                self.logger.info('errno.EWOULDBLOCK')
            else:
                self.logger.error('exception occur. msg[%s], traceback[%s]', str(e), __import__('traceback').format_exc())

    @property
    def logger(self):
        raise Exception('should override')

    def is_ga_request(self, path):
        """
        request是否要被统计
        """

        # 先判断是否在forbid列表里，只要发现就直接拒绝
        for pattern in self._ga_forbid_paths or tuple():
            if re.match(pattern, path):
                self.logger.debug('path is in forbid paths. patten: %s, path: %s', pattern, path)
                return False

        # 改成不为None就有效，这样空列表也是生效的
        if self._ga_allow_paths is not None:
            for pattern in self._ga_allow_paths:
                if re.match(pattern, path):
                    return True
            else:
                self.logger.debug('path is not in allow paths. path: %s', path)
                return False

        return True

    def hack_path(self, path):
        """
        将path替换为需要上报的
        """
        if not self._ga_hack_paths:
            return path

        for src_p, dst_p in self._ga_hack_paths:
            if not re.match(src_p, path):
                continue

            try:
                return re.sub(src_p, dst_p, path)
            except Exception, e:
                self.logger.error('re.sub fail. path: %s, src_p:%s, dst_p: %s, e: %s', path, src_p, dst_p, e)
                return path

        return path


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
用来作为接受google统计上报的agent
通过udp通道。
"""

import json
import functools
import logging
import logging.config
import socket
import threading
import SocketServer
import signal
import sys
import time
import constants


def alloc_ga_obj_by_name(name):
    """
    通过名字生成对应的obj
    """
    from pyga.requests import Tracker, Page, Session, Visitor, Transaction, Event, SocialInteraction, Campaign

    obj = None
    if name == 'tracker':
        obj = Tracker()
    elif name == 'campaign':
        obj = Campaign(Campaign.TYPE_REFERRAL)
    elif name == 'session':
        obj = Session()
    elif name == 'page':
        obj = Page(None)
    elif name == 'visitor':
        obj = Visitor()
    elif name == 'transaction':
        obj = Transaction()
    elif name == 'event':
        obj = Event()
    elif name == 'social_interaction':
        obj = SocialInteraction()

    return obj


def recur_make_ga_obj(name, conf):
    """
    递归生成
    """
    if isinstance(conf, dict) and conf.get('__ga'):
        obj = alloc_ga_obj_by_name(name)
        if not obj:
            return obj

        for attr_name, attr_value in conf.items():
            if attr_name == '__ga':
                continue
            real_value = recur_make_ga_obj(attr_name, attr_value)
            setattr(obj, attr_name, real_value)

        return obj
    else:
        return conf


class ThreadedUDPRequestHandler(SocketServer.BaseRequestHandler):

    def _handle_message(self, message):
        recv_dict = json.loads(message)

        # 就可以直接删除了
        funcname = recv_dict.pop('funcname', None)
        caller = None
        kwargs = dict()

        for name, conf in recv_dict.items():
            obj = recur_make_ga_obj(name, conf)

            if name == 'tracker':
                caller = obj
            else:
                kwargs[name] = obj
        getattr(caller, funcname)(**kwargs)

    def handle(self):
        message = self.request[0]
        self.server.logger.debug("message, len: %s, content: %s", len(message), message)
        try:
            self._handle_message(message)
        except Exception, e:
            self.server.logger.error('exception occur. e: %s', e)


class GAAgent(SocketServer.ThreadingUDPServer):

    def __init__(self, host=None, port=None, logger_name=None):
        # 因为父类继承是用的老风格，所以必须按照下面的方式来写。 不能使用 super(GAAgent, self).__init__
        SocketServer.ThreadingUDPServer.__init__(self,
                                                 (host or constants.GA_AGENT_HOST,
                                                  port or constants.GA_AGENT_PORT),
                                                 ThreadedUDPRequestHandler)
        self.logger = logging.getLogger(logger_name or constants.GA_AGENT_LOGGER_NAME)

    def run(self):
        server_thread = threading.Thread(target=self.serve_forever)
        # Exit the server thread when the main thread terminates
        server_thread.daemon = True
        server_thread.start()

        # 因为daemon设置为true，所以不做while循环会直接退出
        # 而之所以把 daemon 设置为true，是为了防止进程不结束的问题
        while True:
            time.sleep(1)



__version__ = '0.1.32'

from .ga_agent import GAAgent
try:
    from .django_ga import DjangoGA
except:
    pass
try:
    from .flask_ga import FlaskGA
except:
    pass

try:
    from .kola_ga import KolaGA
except:
    pass


#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
启动 ga_agent
"""

import argparse
import os.path as op
import logging
import logging.config

import sys
sys.path.insert(0, '../../')
import sga
from sga import GAAgent, constants


# 日志
# 为了保证邮件只有在正式环境发送
class RequireDebugOrNot(logging.Filter):
    _need_debug = False

    def __init__(self, need_debug, *args, **kwargs):
        super(RequireDebugOrNot, self).__init__(*args, **kwargs)
        self._need_debug = need_debug
        
    def filter(self, record):
        return debug if self._need_debug else not debug


FILE_MODULE_NAME = op.splitext(op.basename(__file__))[0]

LOG_FILE_PATH = "/tmp/ga_agent.log"

LOG_FORMAT = '\n'.join((
    '/' + '-' * 80,
    '[%(levelname)s][%(asctime)s][%(process)d:%(thread)d][%(filename)s:%(lineno)d %(funcName)s]:',
    '%(message)s',
    '-' * 80 + '/',
))

LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,

    'formatters': {
        'standard': {
            'format': LOG_FORMAT,
        },
    },

    'filters': {
        'require_debug_false': {
            '()': RequireDebugOrNot,
            'need_debug': False,
        },
        'require_debug_true': {
            '()': RequireDebugOrNot,
            'need_debug': True,
        },
    },

    'handlers': {
        'rotating_file': {
            'level': 'INFO',
            'class': 'logging.handlers.RotatingFileHandler',
            'formatter': 'standard',
            'filename': LOG_FILE_PATH,
            'maxBytes': 1024 * 1024 * 500,  # 500 MB
            'backupCount': 5,
        },
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'standard',
            'filters': ['require_debug_true'],
        },
    },

    'loggers': {
        'default': {
            'handlers': ['console', 'rotating_file'],
            'level': 'DEBUG',
            'propagate': False
        },
        'ga_agent': {
            'handlers': ['console', 'rotating_file'],
            'level': 'DEBUG',
            'propagate': False
        },
    }
}


logger = logging.getLogger('default')
debug = False


def build_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('-t', '--host', default=constants.GA_AGENT_HOST, help='bind host', action='store')
    parser.add_argument('-p', '--port', default=constants.GA_AGENT_PORT, type=int, help='bind port', action='store')
    parser.add_argument('-d', '--debug', default=False, help='debug mode', action='store_true')
    parser.add_argument('-v', '--version', action='version', version='%s' % sga.__version__)
    return parser


def configure_logging():
    logging.config.dictConfig(LOGGING)

 
def run_ga_agent():
    global debug

    configure_logging()

    args = build_parser().parse_args()

    # 设置到全局配置里
    debug = args.debug

    logger.info("Running GAAgent on %(host)s:%(port)s, debug:%(debug)s" % dict(
        host=args.host, port=args.port, debug=args.debug)
    )

    prog = GAAgent(args.host, args.port)
    try:
        prog.run()
    except KeyboardInterrupt:
        sys.exit(0)
 
if __name__ == '__main__':
    run_ga_agent()



# -*- coding: utf-8 -*-

"""
flask插件，绑定之后可以自动给本地的ga_agent发送数据
"""
import logging
import json
import socket
import errno
import time
import re
import urlparse

from flask import request, g

import constants
from ga_adapter import GAAdapter


class FlaskGA(GAAdapter):

    def __init__(self, app=None):
        super(FlaskGA, self).__init__()

        if app:
            self.init_app(app)

    def init_app(self, app):
        """
        绑定app
        """
        self._ga_id = app.config.get('GA_ID')
        self._ga_agent_host = app.config.get('GA_AGENT_HOST') or constants.GA_AGENT_HOST
        self._ga_agent_port = app.config.get('GA_AGENT_PORT') or constants.GA_AGENT_PORT
        self._ga_forbid_paths = app.config.get('GA_FORBID_PATHS')
        self._ga_allow_paths = app.config.get('GA_ALLOW_PATHS')
        self._ga_hack_paths = app.config.get('GA_HACK_PATHS')
        self._ga_logger_name = app.config.get('GA_LOGGER_NAME')

        @app.before_request
        def prepare_ga_data():
            g.ga_begin_time = time.time()

        @app.teardown_request
        def send_ga_data(exc):
            self.logger.debug('ga_id:%s', self._ga_id)

            if not self._ga_id:
                return False

            if not self.is_ga_request(request.path):
                return False

            try:
                send_dict = self._gen_send_dict()
                if not send_dict:
                    # 这个时候不是正常的请求，比如是用test_request_context模拟的
                    self.logger.debug('invalid request, may be in test_request_context')
                    return False
                self.send_data_to_ga_agent(send_dict)

                return True
            except Exception, e:
                self.logger.exception('exception')

            return False

    @property
    def logger(self):
        return logging.getLogger(self._ga_logger_name or 'flask_ga')

    def _gen_send_dict(self):
        """
        生成发送的dict
        """
        if not getattr(g, 'ga_begin_time', None):
            return None

        load_time = int((time.time()-g.ga_begin_time) * 1000 * 1000)

        ga_referrer_path = ''
        if request.referrer:
            try:
                parse_result = urlparse.urlparse(request.referrer)
                ga_referrer_path = '/%s%s' % (parse_result.netloc, parse_result.path)
            except Exception, e:
                self.logger.info('urlparse fail. e: %s', e)

        send_dict = dict(
            funcname='track_pageview',
            tracker=dict(
                __ga=True,
                account_id=self._ga_id,
                domain_name=request.host,
                campaign=dict(
                    __ga=True,
                    source=self._local_ip,
                    content=ga_referrer_path,
                ),
            ),
            session=dict(
                __ga=True,
            ),
            page=dict(
                __ga=True,
                path=self.hack_path(request.path),
                load_time=load_time,
            ),
            visitor=dict(
                __ga=True,
                ip_address=request.remote_addr,
            ),
        )

        return send_dict



# -*- coding: utf-8 -*-

"""
django插件，绑定之后可以自动给本地的ga_agent发送数据
"""

import logging
import socket
import json
import errno
import time
import re
import urlparse
from django.conf import settings

import constants
from ga_adapter import GAAdapter


class DjangoGA(GAAdapter):

    def __init__(self):
        super(DjangoGA, self).__init__()

        self._ga_id = getattr(settings, 'GA_ID', None)
        self._ga_agent_host = getattr(settings, 'GA_AGENT_HOST', None) or constants.GA_AGENT_HOST
        self._ga_agent_port = getattr(settings, 'GA_AGENT_PORT', None) or constants.GA_AGENT_PORT
        self._ga_forbid_paths = getattr(settings, 'GA_FORBID_PATHS', None)
        self._ga_allow_paths = getattr(settings, 'GA_ALLOW_PATHS', None)
        self._ga_hack_paths = getattr(settings, 'GA_HACK_PATHS', None)
        self._ga_logger_name = getattr(settings, 'GA_LOGGER_NAME', None)

    def process_request(self, request):
        request.ga_begin_time = time.time()

    def process_response(self, request, response):
        """
        无论是否抛出异常，都会执行这一步
        """
        self._send_ga_data(request)
        return response

    @property
    def logger(self):
        return logging.getLogger(self._ga_logger_name or 'django_ga')

    def _send_ga_data(self, request):
        self.logger.debug('ga_id:%s', self._ga_id)

        if not self._ga_id:
            return False

        if not self.is_ga_request(request.path):
            return False

        try:
            send_dict = self._gen_send_dict(request)
            if not send_dict:
                self.logger.debug('invalid request')
                return False
            self.send_data_to_ga_agent(send_dict)

            return True
        except Exception, e:
            self.logger.exception('exception')

        return False

    def _gen_send_dict(self, request):
        """
        生成发送的dict
        """
        if not getattr(request, 'ga_begin_time', None):
            return None

        load_time = int((time.time()-request.ga_begin_time) * 1000 * 1000)

        ga_referrer_path = ''
        if request.META.get('HTTP_REFERER', ''):
            try:
                parse_result = urlparse.urlparse(request.META['HTTP_REFERER'])
                ga_referrer_path = '/%s%s' % (parse_result.netloc, parse_result.path)
            except Exception, e:
                self.logger.info('urlparse fail. e: %s', e)

        send_dict = dict(
            funcname='track_pageview',
            tracker=dict(
                __ga=True,
                account_id=self._ga_id,
                domain_name=request.get_host(),
                campaign=dict(
                    __ga=True,
                    source=self._local_ip,
                    content=ga_referrer_path,
                ),
            ),
            session=dict(
                __ga=True,
            ),
            page=dict(
                __ga=True,
                path=self.hack_path(request.path),
                load_time=load_time,
            ),
            visitor=dict(
                __ga=True,
                ip_address=request.META.get('REMOTE_ADDR', ''),
            ),
        )

        return send_dict



# -*- coding: utf-8 -*-

import time
from ga_adapter import GAAdapter
import constants


class KolaGA(GAAdapter):
    """
    上报GA
    使用:
        KolaGA(dict(...))).register_to_app(app)
    config: dict类型
        GA_ID   必填
        GA_AGENT_HOST
        GA_AGENT_PORT
        GA_FORBID_ENDPOINTS
        GA_ALLOW_ENDPOINTS
    """

    def __init__(self, config):
        GAAdapter.__init__(self)

        self._ga_id = config.get('GA_ID')
        self._ga_agent_host = config.get('GA_AGENT_HOST') or constants.GA_AGENT_HOST
        self._ga_agent_port = config.get('GA_AGENT_PORT') or constants.GA_AGENT_PORT
        self._ga_forbid_paths = config.get('GA_FORBID_PATHS')
        self._ga_allow_paths = config.get('GA_ALLOW_PATHS')
        self._ga_hack_paths = config.get('GA_HACK_PATHS')
        self._ga_logger_name = config.get('GA_LOGGER_NAME')

    def register_to_app(self, app):
        """
        注册相关的函数
        """

        @app.before_app_request
        def prepare_ga_data(request):
            request.ga_begin_time = time.time()

        @app.after_app_request
        def send_ga_data(request, result):
            if not self.is_ga_request(request.values.get('endpoint')):
                return

            try:
                send_dict = self.gen_send_dict(request)
                if not send_dict:
                    # 这个时候不是正常的请求，比如是用test_request_context模拟的
                    self.logger.debug('invalid request')
                    return False
                self.send_data_to_ga_agent(send_dict)

                return True
            except Exception, e:
                self.logger.exception('exception')

            return False

    def gen_send_dict(self, request):
        """
        生成发送的dict
        """
        if not getattr(request, 'ga_begin_time', None):
            return None

        # 为了解决总是0的问题
        load_time = int((time.time()-request.ga_begin_time) * 1000 * 1000)

        send_dict = dict(
            funcname='track_pageview',
            tracker=dict(
                __ga=True,
                account_id=self._ga_id,
                domain_name=self._local_ip,
                campaign=dict(
                    __ga=True,
                    source=self._local_ip,
                    content='',
                ),
            ),
            session=dict(
                __ga=True,
            ),
            page=dict(
                __ga=True,
                path='/' + request.values.get('endpoint'),
                load_time=load_time,
            ),
            visitor=dict(
                __ga=True,
                ip_address=request.address,
            ),
        )

        return send_dict



from setuptools import setup

setup(
    name="sga",
    version='0.1.32',
    zip_safe=False,
    platforms='any',
    packages=['sga'],
    install_requires=["pyga"],
    scripts=['sga/bin/run_ga_agent.py'],
    url="https://github.com/dantezhu/sga",
    license="BSD",
    author="dantezhu",
    author_email="zny2008@gmail.com",
    description="make it easier to use pyga for web develop. and make pyga compatible with flask and django.",
)
